@@ -7,19 +7,41 @@
  */
 #include "engine.h"
 #include <fcntl.h>
+#include <algorithm>
+#include <cstddef>
+#include <cstdint>
+#include <memory>
+#include <set>
+#include <string>
+#include <string_view>
+#include <unordered_map>
+#include <utility>
+#include <vector>
+#include <fcitx-config/iniparser.h>
+#include <fcitx-config/rawconfig.h>
+#include <fcitx-utils/capabilityflags.h>
+#include <fcitx-utils/fs.h>
+#include <fcitx-utils/i18n.h>
 #include <fcitx-utils/inputbuffer.h>
+#include <fcitx-utils/key.h>
+#include <fcitx-utils/keysym.h>
 #include <fcitx-utils/log.h>
+#include <fcitx-utils/misc.h>
+#include <fcitx-utils/standardpath.h>
+#include <fcitx-utils/stringutils.h>
 #include <fcitx-utils/utf8.h>
-#include <keyman/keyboardprocessor.h>
+#include <fcitx/addoninstance.h>
+#include <fcitx/event.h>
+#include <fcitx/inputcontext.h>
+#include <fcitx/inputcontextproperty.h>
+#include <fcitx/inputmethodentry.h>
+#include <fcitx/instance.h>
+#include <keyman_core_api.h>
+#include <keyman_core_api_vkeys.h>
+#include <km_types.h>
 #include "kmpdata.h"
 #include "kmpmetadata.h"
 
-// Keyman introduced this in release-15.0.115-alpha, which is later than
-// km_kbp_process_event API change.
-#if defined(KM_KBP_VERSION_MAJOR)
-#define NEW_KM_KBP_PROCESS_EVENT
-#endif
-
 #define MAXCONTEXT_ITEMS 128
 #define KEYMAN_BACKSPACE 14
 #define KEYMAN_BACKSPACE_KEYSYM 0xff08
@@ -28,14 +50,15 @@
 #define KEYMAN_RCTRL 97
 #define KEYMAN_RALT 100
 
-FCITX_DEFINE_LOG_CATEGORY(keyman, "keyman");
-#define FCITX_KEYMAN_DEBUG() FCITX_LOGC(::keyman, Debug)
-#define FCITX_KEYMAN_ERROR() FCITX_LOGC(::keyman, Error)
+#define FCITX_KEYMAN_DEBUG() FCITX_LOGC(::fcitx::keyman, Debug)
+#define FCITX_KEYMAN_ERROR() FCITX_LOGC(::fcitx::keyman, Error)
 
 namespace fcitx {
 
 namespace {
 
+FCITX_DEFINE_LOG_CATEGORY(keyman, "keyman");
+
 std::vector<char16_t> utf8ToUTF16(std::string_view str) {
     if (!utf8::validate(str)) {
         return {};
@@ -84,29 +107,19 @@ std::string utf16ToUTF8(Iter start, Iter end) {
     return result;
 }
 
-std::string get_current_context_text(km_kbp_context *context) {
+std::string get_current_context_text_debug(km_core_state *state) {
+    km_core_cp *buf =
+        km_core_state_context_debug(state, KM_CORE_DEBUG_CONTEXT_CACHED);
     std::string result;
-    UniqueCPtr<km_kbp_context_item, km_kbp_context_items_dispose>
-        context_items_ptr;
-    km_kbp_status status = KM_KBP_STATUS_OK;
-    {
-        km_kbp_context_item *context_items = nullptr;
-        status = km_kbp_context_get(context, &context_items);
-        context_items_ptr.reset(context_items);
-    }
-    if (status == KM_KBP_STATUS_OK) {
-        size_t buf_size = 0;
-        km_kbp_context_items_to_utf8(context_items_ptr.get(), nullptr,
-                                     &buf_size);
-        if (buf_size) {
-            std::vector<char> buf;
-            buf.resize(buf_size + 1);
-            km_kbp_context_items_to_utf8(context_items_ptr.get(), buf.data(),
-                                         &buf_size);
-            return {buf.data()};
+    if (buf) {
+        size_t n = 0;
+        while (buf[n]) {
+            n++;
         }
+        result = utf16ToUTF8(buf, buf + n);
+        km_core_cp_dispose(buf);
     }
-    return "";
+    return result;
 }
 
 std::set<std::string> listKeymapDirs() {
@@ -129,24 +142,24 @@ class KeymanState : public InputContextProperty {
 public:
     KeymanState(KeymanKeyboardData *keyboard, InputContext *ic)
         : keyboard_(keyboard), ic_(ic) {
-        std::vector<km_kbp_option_item> keyboard_opts;
+        std::vector<km_core_option_item> keyboard_opts;
 
         keyboard_opts.emplace_back();
-        keyboard_opts.back().scope = KM_KBP_OPT_ENVIRONMENT;
+        keyboard_opts.back().scope = KM_CORE_OPT_ENVIRONMENT;
         const auto platform = utf8ToUTF16("platform");
         keyboard_opts.back().key = platform.data();
         const auto platformValue = utf8ToUTF16("linux desktop hardware native");
         keyboard_opts.back().value = platformValue.data();
 
         keyboard_opts.emplace_back();
-        keyboard_opts.back().scope = KM_KBP_OPT_ENVIRONMENT;
+        keyboard_opts.back().scope = KM_CORE_OPT_ENVIRONMENT;
         const auto baseLayout = utf8ToUTF16("baseLayout");
         keyboard_opts.back().key = baseLayout.data();
         const auto baseLayoutValue = utf8ToUTF16("kbdus.dll");
         keyboard_opts.back().value = baseLayoutValue.data();
 
         keyboard_opts.emplace_back();
-        keyboard_opts.back().scope = KM_KBP_OPT_ENVIRONMENT;
+        keyboard_opts.back().scope = KM_CORE_OPT_ENVIRONMENT;
         const auto baseLayoutAlt = utf8ToUTF16("baseLayoutAlt");
         keyboard_opts.back().key = baseLayoutAlt.data();
         const auto baseLayoutAltValue = utf8ToUTF16("en-US");
@@ -156,21 +169,20 @@ public:
         keyboard_opts.back().scope = 0;
         keyboard_opts.back().key = nullptr;
         keyboard_opts.back().value = nullptr;
-        km_kbp_status status_state = km_kbp_state_create(
+        km_core_status status_state = km_core_state_create(
             keyboard_->kbpKeyboard(), keyboard_opts.data(), &state);
-        if (status_state != KM_KBP_STATUS_OK) {
-            FCITX_KEYMAN_ERROR() << "problem creating km_kbp_state for "
+        if (status_state != KM_CORE_STATUS_OK) {
+            FCITX_KEYMAN_ERROR() << "problem creating km_core_state for "
                                  << keyboard_->metadata().id;
             return;
         };
-        resetContext();
+        updateContext();
     }
 
-    void resetContext() {
-        auto context = km_kbp_state_context(state);
-        if (ic_->capabilityFlags().test(CapabilityFlag::SurroundingText)) {
-            auto current_context_utf8 = get_current_context_text(context);
-
+    // Update context from surrounding if possible.
+    void updateContext() {
+        if (ic_->capabilityFlags().test(CapabilityFlag::SurroundingText) &&
+            ic_->surroundingText().isValid()) {
             auto text = ic_->surroundingText().text();
             auto context_pos = std::min(ic_->surroundingText().anchor(),
                                         ic_->surroundingText().cursor());
@@ -182,40 +194,33 @@ public:
             auto endIter =
                 utf8::nextNChar(startIter, context_pos - context_start);
             std::string new_context(startIter, endIter);
-            if (!stringutils::endsWith(new_context, current_context_utf8)) {
-                FCITX_KEYMAN_DEBUG()
-                    << "setting context because it has changed from expected";
-                km_kbp_context_item *context_items = nullptr;
-                if (km_kbp_context_items_from_utf8(new_context.data(),
-                                                   &context_items) ==
-                    KM_KBP_STATUS_OK) {
-                    km_kbp_context_set(context, context_items);
-                }
-                km_kbp_context_items_dispose(context_items);
-            }
-        } else {
-            km_kbp_context_clear(context);
+            auto utf16Context = utf8ToUTF16(new_context);
+            km_core_state_context_set_if_needed(
+                state, reinterpret_cast<km_core_cp *>(utf16Context.data()));
+            FCITX_KEYMAN_DEBUG()
+                << "Set context from application: " << new_context;
         }
     }
 
+    void clearContext() const {
+        FCITX_KEYMAN_DEBUG() << "Clear context";
+        km_core_state_context_clear(state);
+    }
+
     void reset() {
         lctrl_pressed = false;
         rctrl_pressed = false;
         lalt_pressed = false;
         ralt_pressed = false;
-        char_buffer.clear();
-        emitting_keystroke = false;
     }
 
     auto *keyboard() { return keyboard_; }
 
-    km_kbp_state *state = nullptr;
+    km_core_state *state = nullptr;
     bool lctrl_pressed = false;
     bool rctrl_pressed = false;
     bool lalt_pressed = false;
     bool ralt_pressed = false;
-    InputBuffer char_buffer{InputBufferOption::FixedCursor};
-    bool emitting_keystroke = false;
 
 private:
     KeymanKeyboardData *keyboard_;
@@ -278,7 +283,7 @@ std::vector<InputMethodEntry> KeymanEngine::listInputMethods() {
     for (auto &[id, keyboard] : keyboards) {
         std::string icon = "km-config";
         // Check if icon file exists, otherwise fallback to keyman's icon.
-        for (auto *suffix : {".bmp.png", ".icon.png"}) {
+        for (const auto *suffix : {".bmp.png", ".icon.png"}) {
             auto path = stringutils::joinPath(keyboard->baseDir,
                                               stringutils::concat(id, suffix));
             if (fs::isreg(path)) {
@@ -296,10 +301,6 @@ std::vector<InputMethodEntry> KeymanEngine::listInputMethods() {
     return result;
 }
 
-FCITX_ADDON_FACTORY(fcitx::KeymanEngineFactory);
-
-} // namespace fcitx
-
 void fcitx::KeymanKeyboardData::load() {
     if (loaded_) {
         return;
@@ -318,12 +319,12 @@ void fcitx::KeymanKeyboardData::load() {
         return;
     }
 
-    km_kbp_status status_keyboard =
-        km_kbp_keyboard_load(kmxPath.data(), &keyboard_);
+    km_core_status status_keyboard =
+        km_core_keyboard_load(kmxPath.data(), &keyboard_);
 
-    if (status_keyboard != KM_KBP_STATUS_OK) {
+    if (status_keyboard != KM_CORE_STATUS_OK) {
         FCITX_KEYMAN_ERROR()
-            << "problem creating km_kbp_keyboard" << metadata_.id;
+            << "problem creating km_core_keyboard" << metadata_.id;
         return;
     }
 
@@ -336,13 +337,13 @@ void fcitx::KeymanKeyboardData::load() {
     FCITX_KEYMAN_DEBUG() << config_;
 }
 
-void fcitx::KeymanKeyboardData::setOption(const km_kbp_cp *key,
-                                          const km_kbp_cp *value) {
-    auto keyEnd = key;
+void fcitx::KeymanKeyboardData::setOption(const km_core_cp *key,
+                                          const km_core_cp *value) {
+    const auto *keyEnd = key;
     while (*keyEnd) {
         ++keyEnd;
     }
-    auto valueEnd = value;
+    const auto *valueEnd = value;
     while (*valueEnd) {
         ++valueEnd;
     }
@@ -366,27 +367,20 @@ fcitx::KeymanKeyboardData::KeymanKeyboardData(
 fcitx::KeymanKeyboardData::~KeymanKeyboardData() { factory_.unregister(); }
 
 void fcitx::KeymanEngine::activate(const fcitx::InputMethodEntry &entry,
-                                   fcitx::InputContextEvent &) {
-    auto data = static_cast<const KeymanKeyboard *>(entry.userData());
+                                   fcitx::InputContextEvent &event) {
+    const auto *data = static_cast<const KeymanKeyboard *>(entry.userData());
     data->load();
-}
-
-static bool ok_for_single_backspace(const km_kbp_action_item *action_items,
-                                    size_t i, size_t num_actions) {
-    for (size_t j = i + 1; j < num_actions; j++) {
-        if (action_items[j].type == KM_KBP_IT_BACK ||
-            action_items[j].type == KM_KBP_IT_CHAR ||
-            action_items[j].type == KM_KBP_IT_EMIT_KEYSTROKE) {
-            return false;
-        }
+    auto *keyman = state(entry, *event.inputContext());
+    if (!keyman) {
+        return;
     }
-    return true;
+    keyman->updateContext();
 }
 
 void fcitx::KeymanEngine::keyEvent(const fcitx::InputMethodEntry &entry,
                                    fcitx::KeyEvent &keyEvent) {
-    auto ic = keyEvent.inputContext();
-    auto keyman = state(entry, *ic);
+    auto *ic = keyEvent.inputContext();
+    auto *keyman = state(entry, *ic);
     if (!keyman) {
         return;
     }
@@ -408,213 +402,135 @@ void fcitx::KeymanEngine::keyEvent(const fcitx::InputMethodEntry &entry,
     default:
         break;
     }
-#ifndef NEW_KM_KBP_PROCESS_EVENT
-    if (keyEvent.isRelease()) {
-        return;
-    }
-#endif
 
     if (keycode < 0 || keycode > 255) {
         return;
     }
 
     if (keycode_to_vk[keycode] == 0) {
-        // key we don't handles
+        // key that we don't handles
+        if (keyEvent.key().isCursorMove()) {
+            km_core_state_context_clear(keyman->state);
+            keyman->updateContext();
+        }
         return;
     }
 
     // keyman modifiers are different from X11
     uint16_t km_mod_state = 0;
     if (state.test(KeyState::Shift)) {
-        km_mod_state |= KM_KBP_MODIFIER_SHIFT;
+        km_mod_state |= KM_CORE_MODIFIER_SHIFT;
     }
     if (state.test(KeyState::Mod5)) {
-        km_mod_state |= KM_KBP_MODIFIER_RALT;
-        FCITX_KEYMAN_DEBUG() << "modstate KM_KBP_MODIFIER_RALT from Mod5";
+        km_mod_state |= KM_CORE_MODIFIER_RALT;
+        FCITX_KEYMAN_DEBUG() << "modstate KM_CORE_MODIFIER_RALT from Mod5";
     }
     if (state.test(KeyState::Mod1)) {
         if (keyman->ralt_pressed) {
-            km_mod_state |= KM_KBP_MODIFIER_RALT;
+            km_mod_state |= KM_CORE_MODIFIER_RALT;
             FCITX_KEYMAN_DEBUG()
-                << "modstate KM_KBP_MODIFIER_RALT from ralt_pressed";
+                << "modstate KM_CORE_MODIFIER_RALT from ralt_pressed";
         }
         if (keyman->lalt_pressed) {
-            km_mod_state |= KM_KBP_MODIFIER_LALT;
+            km_mod_state |= KM_CORE_MODIFIER_LALT;
             FCITX_KEYMAN_DEBUG()
-                << "modstate KM_KBP_MODIFIER_LALT from lalt_pressed";
+                << "modstate KM_CORE_MODIFIER_LALT from lalt_pressed";
         }
     }
     if (state.test(KeyState::Ctrl)) {
         if (keyman->rctrl_pressed) {
-            km_mod_state |= KM_KBP_MODIFIER_RCTRL;
+            km_mod_state |= KM_CORE_MODIFIER_RCTRL;
             FCITX_KEYMAN_DEBUG()
-                << "modstate KM_KBP_MODIFIER_RCTRL from rctrl_pressed";
+                << "modstate KM_CORE_MODIFIER_RCTRL from rctrl_pressed";
         }
         if (keyman->lctrl_pressed) {
-            km_mod_state |= KM_KBP_MODIFIER_LCTRL;
+            km_mod_state |= KM_CORE_MODIFIER_LCTRL;
             FCITX_KEYMAN_DEBUG()
-                << "modstate KM_KBP_MODIFIER_LCTRL from lctrl_pressed";
+                << "modstate KM_CORE_MODIFIER_LCTRL from lctrl_pressed";
         }
     }
-    auto context = km_kbp_state_context(keyman->state);
+
+    if (ic->capabilityFlags().test(CapabilityFlag::SurroundingText) &&
+        ic->surroundingText().isValid()) {
+        keyman->updateContext();
+    }
+
     FCITX_KEYMAN_DEBUG() << "before process key event context: "
-                         << get_current_context_text(context);
+                         << get_current_context_text_debug(keyman->state);
     FCITX_KEYMAN_DEBUG() << "km_mod_state=" << km_mod_state;
-#ifdef NEW_KM_KBP_PROCESS_EVENT
-    km_kbp_process_event(keyman->state, keycode_to_vk[keycode], km_mod_state,
-                         !keyEvent.isRelease());
-#else
-    km_kbp_process_event(keyman->state, keycode_to_vk[keycode], km_mod_state);
-#endif
+    km_core_process_event(keyman->state, keycode_to_vk[keycode], km_mod_state,
+                          !keyEvent.isRelease(), 0);
+    FCITX_KEYMAN_DEBUG() << "after process key event context : "
+                         << get_current_context_text_debug(keyman->state);
 
-    FCITX_KEYMAN_DEBUG() << "after process key event context: "
-                         << get_current_context_text(context);
+    km_core_actions const *actions = km_core_state_get_actions(keyman->state);
+    auto numOfDelete = actions->code_points_to_delete;
+    FCITX_KEYMAN_DEBUG() << "BACK action " << numOfDelete;
 
-    // km_kbp_state_action_items to get action items
-    size_t num_action_items;
-    const km_kbp_action_item *action_items =
-        km_kbp_state_action_items(keyman->state, &num_action_items);
-
-    std::vector<uint16_t> utf16Buffer;
-    auto flushUtf16Buffer = [&utf16Buffer, keyman]() {
-        if (utf16Buffer.empty()) {
-            return;
-        }
-        std::string utf8 = utf16ToUTF8(utf16Buffer.begin(), utf16Buffer.end());
-        keyman->char_buffer.type(utf8);
-        utf16Buffer.clear();
-    };
-
-    for (size_t i = 0; i < num_action_items; i++) {
-        if (action_items[i].type == KM_KBP_IT_CHAR) {
+    if (numOfDelete > 0) {
+        if (numOfDelete == 1 && keyEvent.key().check(FcitxKey_BackSpace)) {
+            emit_keystroke = true;
+        } else if (ic->capabilityFlags().test(
+                       CapabilityFlag::SurroundingText)) {
+            ic->deleteSurroundingText(-numOfDelete, numOfDelete);
             FCITX_KEYMAN_DEBUG()
-                << "CHAR action " << i + 1 << "/" << num_action_items;
-            utf16Buffer.push_back(action_items[i].character);
-            continue;
+                << "deleting surrounding text " << numOfDelete << " char(s)";
         } else {
-            flushUtf16Buffer();
-        }
-
-        switch (action_items[i].type) {
-        case KM_KBP_IT_MARKER:
-            FCITX_KEYMAN_DEBUG()
-                << "MARKER action " << i + 1 << "/" << num_action_items;
-            break;
-        case KM_KBP_IT_ALERT:
-            FCITX_KEYMAN_DEBUG()
-                << "ALERT action " << i + 1 << "/" << num_action_items;
-            break;
-        case KM_KBP_IT_BACK:
-            FCITX_KEYMAN_DEBUG()
-                << "BACK action " << i + 1 << "/" << num_action_items;
-            if (!keyman->char_buffer.empty()) {
-                keyman->char_buffer.backspace();
-            } else if (ok_for_single_backspace(action_items, i,
-                                               num_action_items)) {
-                FCITX_KEYMAN_DEBUG() << "no char actions, just single back";
-                return;
-            } else {
-                if (ic->capabilityFlags().test(
-                        CapabilityFlag::SurroundingText)) {
-                    ic->deleteSurroundingText(-1, 1);
-                    FCITX_KEYMAN_DEBUG() << "deleting surrounding text 1 char";
-                } else {
-                    FCITX_KEYMAN_DEBUG()
-                        << "forwarding backspace with reset context";
-                    km_kbp_context_item *context_items;
-                    km_kbp_context_get(km_kbp_state_context(keyman->state),
-                                       &context_items);
-                    keyman->resetContext();
-                    ic->forwardKey(Key(FcitxKey_BackSpace));
-                    km_kbp_context_set(km_kbp_state_context(keyman->state),
-                                       context_items);
-                    km_kbp_context_items_dispose(context_items);
-                }
+            FCITX_KEYMAN_DEBUG() << "forwarding backspace with reset context";
+            while (numOfDelete) {
+                ic->forwardKey(Key(FcitxKey_BackSpace));
+                numOfDelete -= 1;
             }
-            break;
-        case KM_KBP_IT_PERSIST_OPT:
-            FCITX_KEYMAN_DEBUG()
-                << "PERSIST_OPT action " << i + 1 << "/" << num_action_items;
-            // Save keyboard option
-            if (action_items[i].option != NULL) {
-                // Allocate for 1 option plus 1 pad struct of 0's
-                std::array<km_kbp_option_item, 2> keyboard_opts;
-                memmove(&(keyboard_opts[0]), action_items[i].option,
-                        sizeof(km_kbp_option_item));
-                memset(&keyboard_opts[1], 0, sizeof(km_kbp_option_item));
-                // Propagate to all state.
-                instance_->inputContextManager().foreach([this, &entry,
-                                                          &keyboard_opts](
-                                                             InputContext *ic) {
-                    if (auto keyman = this->state(entry, *ic)) {
-                        auto event_status = km_kbp_state_options_update(
-                            keyman->state, keyboard_opts.data());
-                        if (event_status != KM_KBP_STATUS_OK) {
-                            FCITX_KEYMAN_DEBUG()
-                                << "problem saving option for km_kbp_keyboard";
-                        }
-                    }
-                    return true;
-                });
-
-                // Put the keyboard option into config
-                if (action_items[i].option != NULL &&
-                    action_items[i].option->key != NULL &&
-                    action_items[i].option->value != NULL) {
-                    FCITX_KEYMAN_DEBUG() << "Saving keyboard option to Config";
-                    // Load the current keyboard options from DConf
-                    keyman->keyboard()->setOption(
-                        action_items[i].option->key,
-                        action_items[i].option->value);
-                }
-            }
-            break;
-        case KM_KBP_IT_EMIT_KEYSTROKE:
-            if (!keyman->char_buffer.empty()) {
-                ic->commitString(keyman->char_buffer.userInput());
-                keyman->char_buffer.clear();
-            }
-            FCITX_KEYMAN_DEBUG()
-                << "EMIT_KEYSTROKE action " << i + 1 << "/" << num_action_items;
-            keyman->emitting_keystroke = true;
-            break;
-        case KM_KBP_IT_INVALIDATE_CONTEXT:
-            FCITX_KEYMAN_DEBUG() << "INVALIDATE_CONTEXT action " << i + 1 << "/"
-                                 << num_action_items;
-            km_kbp_context_clear(km_kbp_state_context(keyman->state));
-            keyman->resetContext();
-            break;
-        case KM_KBP_IT_END:
-            FCITX_KEYMAN_DEBUG()
-                << "END action " << i + 1 << "/" << num_action_items;
-            if (!keyman->char_buffer.empty()) {
-                ic->commitString(keyman->char_buffer.userInput());
-                keyman->char_buffer.clear();
-            }
-            if (keyman->emitting_keystroke) {
-                keyman->emitting_keystroke = false;
-                return;
-            }
-            break;
-        default:
-            FCITX_KEYMAN_DEBUG()
-                << "Unknown action " << i + 1 << "/" << num_action_items;
         }
     }
-    flushUtf16Buffer();
-    context = km_kbp_state_context(keyman->state);
+
+    std::string output;
+    if (actions->output) {
+        for (size_t n = 0; actions->output[n]; n++) {
+            output.append(utf8::UCS4ToUTF8(actions->output[n]));
+        }
+    }
+
+    if (actions->do_alert) {
+        FCITX_KEYMAN_DEBUG() << "ALERT action";
+    }
+
+    if (!output.empty()) {
+        ic->commitString(output);
+    }
+    if (actions->emit_keystroke || emit_keystroke) {
+        FCITX_KEYMAN_DEBUG() << "EMIT_KEYSTROKE action";
+        emit_keystroke = false;
+    } else {
+        keyEvent.filterAndAccept();
+    }
+
+    FCITX_KEYMAN_DEBUG() << "PERSIST_OPT action";
+    for (size_t i = 0; actions->persist_options[i].scope; i++) {
+        if (actions->persist_options[i].key &&
+            actions->persist_options[i].value) {
+            // Put the keyboard option into config
+            FCITX_KEYMAN_DEBUG() << "Saving keyboard option to Config";
+            // Load the current keyboard options from DConf
+            keyman->keyboard()->setOption(actions->persist_options[i].key,
+                                          actions->persist_options[i].value);
+        }
+    }
+
+    // TODO: set capslock if actions->new_caps_lock_state !=
+    // KM_CORE_CAPS_UNCHANGED
+
     FCITX_KEYMAN_DEBUG() << "after processing all actions";
-    keyEvent.filterAndAccept();
 }
 
 void fcitx::KeymanEngine::reset(const fcitx::InputMethodEntry &entry,
                                 fcitx::InputContextEvent &event) {
-    auto ic = event.inputContext();
-    auto keyman = state(entry, *ic);
+    auto *ic = event.inputContext();
+    auto *keyman = state(entry, *ic);
     if (!keyman) {
         return;
     }
-    keyman->resetContext();
+    keyman->clearContext();
     keyman->reset();
 }
 
@@ -622,13 +538,14 @@ fcitx::KeymanState *
 fcitx::KeymanEngine::state(const fcitx::InputMethodEntry &entry,
                            fcitx::InputContext &ic) {
 
-    auto userData = static_cast<const KeymanKeyboard *>(entry.userData());
+    const auto *userData =
+        static_cast<const KeymanKeyboard *>(entry.userData());
     auto &data = userData->data();
     // Check if data is ready.
     if (!data.kbpKeyboard() || !data.factory().registered()) {
         return nullptr;
     }
-    auto keyman = ic.propertyFor(&data.factory());
+    auto *keyman = ic.propertyFor(&data.factory());
     if (!keyman->state) {
         return nullptr;
     }
@@ -637,9 +554,13 @@ fcitx::KeymanEngine::state(const fcitx::InputMethodEntry &entry,
 
 std::string fcitx::KeymanEngine::subMode(const fcitx::InputMethodEntry &entry,
                                          fcitx::InputContext &ic) {
-    auto keyman = state(entry, ic);
+    auto *keyman = state(entry, ic);
     if (!keyman) {
         return _("Not available");
     }
     return "";
 }
+
+} // namespace fcitx
+
+FCITX_ADDON_FACTORY_V2(keyman, fcitx::KeymanEngineFactory);
